#include "SpaceSubwindow.h"

#include "Camera.h"
#include "Earth.h"
#include "Globals.h"
#include "Keys.h"
#include "Pass.h"
#include "Orbit.h"
#include "OpenGLFunctions.h"
#include "Screen.h"
#include "StarField.h"
#include "Textures.h"
#include "Utils.h"

#include <GL/glut.h>
#include <math.h>
#include <stdio.h>

//Prototypes
void spaceSetUpView();
void spaceSetUpModelTransform();
void spaceSetUpLight();
void drawSun();
void drawPassDateTimes();
void drawCurrentGlyphsDisplayed();

#define HUD_X_START 35



#define SPACE_WIDTH 10000
#define SPACE_HEIGHT 10000
#define SPACE_DEPTH 10000

#define SUN_X -50000
#define SUN_Y 100000
#define SUN_Z 10000
#define SUN_RADIUS 1000
#define SUN_RESOLUTION 100

// Orbit params
#define INCLINATION 98.0207
//#define ALTITUDE 700
#define ALTITUDE 150
#define PERIOD 150

#define STEP_DISTANCE 5

#define Z_DISTANCE -400


static int normalFont = (int) GLUT_BITMAP_8_BY_13;
static int headerFont = (int) GLUT_BITMAP_TIMES_ROMAN_24;
static int subheaderFont = (int) GLUT_BITMAP_HELVETICA_18;

// Global File Vars
Screen *spaceScreen;

Pass *pass;
Earth *earth;
Orbit *orbit;
StarField *starField;

GLUquadricObj* sunSphere;

static CameraMode cameraMode = CAMERA_MODE_GLOBAL;

char showStars = 1, showEarth = 1;

// View variables
static GLfloat modelTrans[] = {0.0, 0.0, Z_DISTANCE};

#define ORG_THETA_X -90.0
#define ORG_THETA_Y 0.0
#define ORG_THETA_Z 0.0

// In degrees
static GLfloat theta[] = {ORG_THETA_X, ORG_THETA_Y, ORG_THETA_Z};
static float thetaIncr = 5.0;

// animation transform variables
//static GLfloat translate[3] = {0.0, 0.0, 0.0};


void resetSpaceCameraCoordinates() {
    if(cameraMode == CAMERA_MODE_GLOBAL){
        // Global
        modelTrans[0] = 0;
        modelTrans[1] = 0;
        modelTrans[2] = Z_DISTANCE;
        
        theta[0] = ORG_THETA_X;
        theta[1] = ORG_THETA_Y;
        theta[2] = ORG_THETA_Z;
    } else {
        // Tracking
        //printf("Last, %f, %f\n",lastOrbitXPos,lastOrbitYPos);
        
        modelTrans[0] = -(orbit->altitude * cos(lastOrbitRotAngle)) ;
        modelTrans[1] = -(orbit->altitude * sin(lastOrbitRotAngle)) ;
        modelTrans[2] = 0.0;
        
        theta[0] = 0.0;
        theta[1] = 0.0;
        theta[2] = 90.0;
    }
}

void spaceSetUpView() {
    // this code initializes the viewing transform
    glLoadIdentity();
    // moves viewer along coordinate axes
    
    if(cameraMode == CAMERA_MODE_TRACKING){
        gluLookAt(-1.0, 0.0, 5.0, 0.0, 0.0, 0.0, -90.0, 0.0, 0.0);
    }
    else {
        gluLookAt(0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    }
    
    // move the view back some relative to viewer[] position
    glTranslatef(0.0f,0.0f, -8.0f);
    // rotates view
    
    if(cameraMode == CAMERA_MODE_TRACKING){
        glRotatef(0.0, 1.0, 0.0, 0.0);
        glRotatef(0.0, 0.0, 1.0, 0.0);
        glRotatef(0.0, 0.0, 0.0, 1.0);
    } else {
        glRotatef(0, 1.0, 0.0, 0.0);
        glRotatef(0, 0.0, 1.0, 0.0);
        glRotatef(0, 0.0, 0.0, 1.0);
    }
    
    return;
}

/*  Function name: setUpModelTransform()
    Description:   Sets up the initial model transform.
*/
void spaceSetUpModelTransform() {
    if(cameraMode == CAMERA_MODE_TRACKING) {
        //resetSpaceCameraCoordinates();
    }
    
   // moves model along coordinate axes
   glTranslatef(modelTrans[0], modelTrans[1], modelTrans[2]);

   // rotates model
   glRotatef(theta[0], 1.0, 0.0, 0.0);
   glRotatef(theta[1], 0.0, 1.0, 0.0);
   glRotatef(theta[2], 0.0, 0.0, 1.0);
}

/* Function name: setUpLight()
   Description:   Sets up the light source.
*/
void spaceSetUpLight() {
    // set up the light sources for the scene
    // a directional light source from directly behind
    /*GLfloat lightDir[] = {0.0, 0.5, 1.0, 0.0};
    GLfloat diffuseComp[] = {1.0, 1.0, 1.0, 1.0};

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    glLightfv(GL_LIGHT0, GL_POSITION, lightDir);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseComp);*/
   
    //GLfloat lightDir[] = {0.0, 0.5, 1.0, 0.0};
    GLfloat lightDir[] = {0.0, 0.5, 1.0, 0.0};
    GLfloat diffuseComp[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat ambientLight[] = {0.1f, 0.1f, 0.1f, 1.0f};
    GLfloat diffuseLight[] = {0.5f, 0.5f, 0.5f, 1.0f};

    glEnable(GL_LIGHTING);
    
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable ( GL_COLOR_MATERIAL ) ;

    glLightfv(GL_LIGHT0, GL_AMBIENT, ambientLight);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
    glEnable(GL_LIGHT0);

    // Old  


    glEnable(GL_LIGHT1);

    glLightfv(GL_LIGHT1, GL_POSITION, lightDir);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuseComp);
    
    

    return;
}


void spaceDisplay (void) {
        
    // this code executes whenever the window is redrawn (when opened,
    // moved, resized, etc.
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // set the viewing transform
    spaceSetUpView();

    // set up light source
    spaceSetUpLight();

    // start drawing objects
    spaceSetUpModelTransform();
    
    // Draw earth
    if(showEarth == 1) {
        drawEarth(earth,0,0,0);
    }
    
    // Draw star field
    if(showStars == 1) {
        drawStarField(starField);
    }
    
    // Draw orbit
    drawOrbit(orbit, 0,0,0);
    
    // Draw sun
    drawSun();
    
    // Print out pass date times
    drawPassDateTimes();
    
    // Draw glyphs displayed
    drawCurrentGlyphsDisplayed();

    glEnd();

   glutSwapBuffers();
}

void spaceSubwindowInit(Pass *passIn) {
    spaceScreen = createScreen();
    
    pass = passIn;
    
    // Create earth
    earth = createEarth(pass);
    
    // Create star field
    starField = createStarField(SPACE_HEIGHT, SPACE_WIDTH, SPACE_DEPTH);
    // printStarField(starField);
    
    // Create Orbit
    orbit = createOrbit(pass, INCLINATION, ALTITUDE, PERIOD);
    indexOfColorGlyph = -1;
    indexOfSizeGlyph = 7;
       
    // Create sun
    sunSphere = gluNewQuadric();
    gluQuadricDrawStyle(sunSphere, GLU_FILL);
    //make opengl generate the texture coordinates for the sphere
    gluQuadricTexture(sunSphere, GL_TRUE);
}

void spaceMouseCallback(int button, int state, int x, int y) {
    // rotate camera
    GLint axis = 3;
    if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) axis = 0;
    if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) axis = 1;
    if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) axis = 2;
    if (axis < 3) {  // button ups won't change axis value from 3
        theta[axis] += thetaIncr;
        if (theta[axis] > 360.0) theta[axis] -= 360.0;
        glutPostRedisplay();
    }
    
    //printf("Current Traslation: %f, %f, %f\n", theta[0], theta[1], theta[2]);
}

void spaceKeyCallback(unsigned char key, int x, int y) {
       // move viewer with x, y, and z keys
       // capital moves in + direction, lower-case - direction

    if (key == 'o' || key == 'O') {
        cycleOrbitDrawType();
    }
    else if (key == KEY_MOVE_X_LEFT) modelTrans[0] -= STEP_DISTANCE;
    else if (key == KEY_MOVE_X_RIGHT) modelTrans[0] += STEP_DISTANCE;
    else if (key == KEY_MOVE_Y_DOWN) modelTrans[1] -= STEP_DISTANCE;
    else if (key == KEY_MOVE_Y_UP) modelTrans[1] += STEP_DISTANCE;
    else if (key == KEY_MOVE_Z_FORWARD) modelTrans[2] -= STEP_DISTANCE;
    else if (key == KEY_MOVE_Z_BACK) modelTrans[2] += STEP_DISTANCE;
    else if (key == KEY_TOGGLE_EARTH_LOWER || key == KEY_TOGGLE_EARTH_UPPER) {
        if(showEarth == 1){
            showEarth = 0;
        }
        else {
            showEarth = 1;
        }
    }
    else if (key == KEY_TOGGLE_STARS_LOWER || key == KEY_TOGGLE_STARS_UPPER) {
        if(showStars == 1){
            showStars = 0;
        }
        else {
            showStars = 1;
        }
    } else if (key == KEY_TOGGLE_ORBIT_CAM) {
        if(cameraMode == CAMERA_MODE_GLOBAL){
            cameraMode = CAMERA_MODE_TRACKING;
        }
        else {
            cameraMode = CAMERA_MODE_GLOBAL;
        }
        resetSpaceCameraCoordinates();
    }
    
    
    
    /*else if (key == 'r') {
       theta[0] = ORG_THETA_X; theta[1] = ORG_THETA_Y; theta[2] = ORG_THETA_Z;
    }
    else if (key == '-') {
       thetaIncr = -thetaIncr;
    }
    else if (key == '+') {
        if (thetaIncr < 0) thetaIncr = thetaIncr - 1.0;
        else               thetaIncr = thetaIncr + 1.0;
    }
    
    else if (key == 't') {
        theta[0] += 1;
    }
    else if (key == 'T') {
        theta[0] -= 1;
    }*/

    glutPostRedisplay();
}


void spaceChangeSize(int w1, int h1) {
	double ratio;

    spaceScreen->width  = w1;
    spaceScreen->height = h1;

    // Prevent a divide by zero, when window is too short
    // (you cant make a window of zero width).
    if(h1 == 0)
        h1 = 1;
    
    ratio = 1.0f * w1 / h1;
    // Reset the coordinate system before modifying
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    // Set the viewport to be the entire window
    glViewport(0, 0, w1, h1);

    // Set the clipping volume
    gluPerspective(45,ratio,0.1,1000000);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void spaceReshapeCallback(int w, int h) {
    // from Angel, p.562
    glViewport(0,0,w,h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (w < h) {
        glFrustum(-2.0, 2.0, -2.0*(GLfloat) h / (GLfloat) w,
                2.0*(GLfloat) h / (GLfloat) w, 2.0, 20.0);
    }
    else {
        glFrustum(-2.0, 2.0, -2.0*(GLfloat) w / (GLfloat) h,
                2.0*(GLfloat) w / (GLfloat) h, 2.0, 20.0);
    }

    glMatrixMode(GL_MODELVIEW);
}

void drawSun() {
    //printf("Drawing sun\n");
    glPushMatrix();

        glTranslatef(SUN_X,SUN_Y,SUN_Z);
        
        glColor3f(1.0,1.0,0);
        
        glEnable(GL_TEXTURE_2D);
        
        glBindTexture(GL_TEXTURE_2D, TEXTURE_SUN);
        gluSphere(sunSphere, SUN_RADIUS, SUN_RESOLUTION, SUN_RESOLUTION);
        
        glDisable(GL_TEXTURE_2D);
    
    glPopMatrix();
}

#define HUD_TITLE_Y_START 40
#define HUD_PASS_RANGE_Y_START 80
#define HUD_PASS_CUR_Y_START 140
#define HUD_GLYPH_Y_START 200

#define HUD_ORBIT_DRAW_Y_START 280
#define HUD_CAMERA_Y_START 340


void drawPassDateTimes() {
    char tempStr[1000];
    
    glDisable(GL_LIGHTING);
    
    glColor3f(1.0,1.0,1.0);
    
    setOrthographicProjection(spaceScreen);
    glPushMatrix();
    glLoadIdentity(); 
    
    renderBitmapString(HUD_X_START, HUD_TITLE_Y_START, (void *)headerFont, "Satellite Data Visualization");
    
    renderBitmapString(HUD_X_START, HUD_PASS_RANGE_Y_START, (void *)subheaderFont, "Pass Time Range");
    renderBitmapString(HUD_X_START, HUD_PASS_RANGE_Y_START+20, (void *)normalFont, pass->dateTimeStr);


    //Current pass time
    renderBitmapString(HUD_X_START, HUD_PASS_CUR_Y_START, (void *)subheaderFont, "Current Pass Time");
    printDateTimeToArray(pass->fields[0]->data[curSelectedIndex]->dateTime, tempStr);
    
    renderBitmapString(HUD_X_START, HUD_PASS_CUR_Y_START+20, (void *)normalFont, tempStr);
    /*
    //Camera mode
    renderBitmapString(HUD_X_START, HUD_CAMERA_Y_START, (void *)subheaderFont, "Camera Mode");
    
    // Print camera
    if(cameraMode == CAMERA_MODE_GLOBAL) {
        renderBitmapString(HUD_X_START, HUD_CAMERA_Y_START + 20, (void *)normalFont, "Global");
    } else {
        renderBitmapString(HUD_X_START, HUD_CAMERA_Y_START + 20, (void *)normalFont, "Tracking");
    }
    */
    // Orbit type
    renderBitmapString(HUD_X_START,HUD_ORBIT_DRAW_Y_START,(void *)subheaderFont, "Graph Type");
    
    if(orbitDrawingMode == ORBIT_DRAW_BAR) {
        renderBitmapString(HUD_X_START,HUD_ORBIT_DRAW_Y_START+20,(void *)normalFont, "Bar");
    }
    if(orbitDrawingMode == ORBIT_DRAW_SPHERE) {
        renderBitmapString(HUD_X_START,HUD_ORBIT_DRAW_Y_START+20,(void *)normalFont, "Sphere");
    }
    
    

    glPopMatrix();
    resetPerspectiveProjection();
    
    glEnable(GL_LIGHTING);
}


void drawCurrentGlyphsDisplayed() {
    char tempStr[1000];
    
    glDisable(GL_LIGHTING);
    
    glColor3f(1.0,1.0,1.0);
    
    setOrthographicProjection(spaceScreen);
    glPushMatrix();
    glLoadIdentity(); 

    renderBitmapString(HUD_X_START, HUD_GLYPH_Y_START, (void *)subheaderFont, "Glyphs Displayed");

    if(indexOfColorGlyph == -1){
        sprintf(tempStr, "Color: (not selected)");
    } else {
        sprintf(tempStr, "Color: %s", pass->fields[indexOfColorGlyph]->name);
    }
    renderBitmapString(HUD_X_START, HUD_GLYPH_Y_START + 20, (void *)normalFont, tempStr);
        
    if(indexOfSizeGlyph == -1){
        sprintf(tempStr, "Size: (not selected)");
    } else {
        sprintf(tempStr, "Size:  %s", pass->fields[indexOfSizeGlyph]->name);
    }
    renderBitmapString(HUD_X_START, HUD_GLYPH_Y_START + 40, (void *)normalFont, tempStr);


    glPopMatrix();
    resetPerspectiveProjection();
    
    glEnable(GL_LIGHTING);
}
